Imports UvARescue.Communication
Imports UvARescue.Math
Imports UvARescue.Tools

Imports System.Threading

Public Class CommActor
    Inherits Actor
    Implements ICommOwner

    Public Shared ReadOnly ACTORTYPE_COMM As String = "Comm"

    'Protected Shared ReadOnly PATHLOSS_THRESHOLD As Single = -57.35 'dBm
    Protected Shared ReadOnly PATHLOSS_THRESHOLD As Single = -93 'dBm, wss cutoff 
    Protected Shared ReadOnly RESYNC_TIMEOUT As Integer = 1 'secs, time to wait to trigger next sync

#Region " Constructor "

    Public Sub New(ByVal teamConfig As TeamConfig, ByVal agentName As String, ByVal agentNumber As Integer, ByVal agentHost As String)
        MyBase.New(ACTORTYPE_COMM, ACTORTYPE_COMM)

        Me._CommDevice = New WssDevice(Me, teamConfig.WirelessServerHost, teamConfig.WirelessServerPort, teamConfig.OperatorName, teamConfig.TeamMembers, agentName, agentNumber, agentHost)

        Me._OperatorName = teamConfig.OperatorName
        Me._TeamMembers = New List(Of String)

        Dim members() As String = Strings.Split(teamConfig.TeamMembers, ",")
        For Each member As String In members
            Dim parts() As String = Strings.Split(member, "-")
            Dim name As String = parts(1)
            Me._TeamMembers.Add(name)
        Next

        ReDim Me._PathLossMatrix(Me._TeamMembers.Count - 1, Me._TeamMembers.Count - 1)
        For i As Integer = 0 To Me._PathLossMatrix.GetLength(0) - 1
            For j As Integer = 0 To Me._PathLossMatrix.GetLength(1) - 1
                If i = j Then
                    Me._PathLossMatrix(i, j) = Integer.MinValue
                End If
            Next
        Next

    End Sub

#End Region

#Region " CommDevice "

    Private _CommDevice As ICommDevice
    Public ReadOnly Property CommDevice() As ICommDevice
        Get
            Return Me._CommDevice
        End Get
    End Property

    Public Sub StartComm()
        Me.CommDevice.StartDevice()
    End Sub

    Public Sub StopComm()
        Me.CommDevice.StopDevice()
    End Sub

#End Region

#Region " Helper Vars "

    Private _AgentName As String
    Private _AgentIndex As Integer

    Private _Manifold As Manifold


    Protected Overrides Sub OnAgentChanged()
        MyBase.OnAgentChanged()
        If Not IsNothing(Me.Agent) Then

            Me._AgentName = Me.Agent.Name
            Me._AgentIndex = Me._TeamMembers.IndexOf(Me._AgentName)
            Me._Manifold = Me.Agent.Manifold

            If TypeOf Me.Agent Is OperatorAgent Then
                Me._Operator = DirectCast(Me.Agent, OperatorAgent)
            Else
                Me._Operator = Nothing
            End If

        Else
            Me._AgentName = String.Empty
            Me._AgentIndex = -1
            Me._Operator = Nothing
            Me._Manifold = Nothing

        End If
    End Sub

    Private _Operator As OperatorAgent = Nothing
    Protected ReadOnly Property [Operator]() As OperatorAgent
        Get
            Return Me._Operator
        End Get
    End Property

#End Region

#Region " Monitor Signal Strength to Team Members "

    Private _OperatorName As String
    Private _TeamMembers As List(Of String)
    Private _PathLossMatrix(,) As Double

    Public Sub NotifySignalStrengthReceived(ByVal toRobot As String, ByVal pathloss As Double) Implements Communication.ICommOwner.NotifySignalStrengthReceived
        Dim j As Integer = Me._TeamMembers.IndexOf(toRobot)


        SyncLock Me._PathLossMatrix
            Me._PathLossMatrix(Me._AgentIndex, j) = pathloss
            Me._PathLossMatrix(j, Me._AgentIndex) = pathloss
        End SyncLock

        If pathloss > PATHLOSS_THRESHOLD AndAlso Not Me.IsConnectedTo(toRobot) Then
            'to avoid each of both agents to open a connection, only the 
            'master will request the conversation

            If Me.IsMasterTo(toRobot) Then
                Me._CommDevice.RequestConversation(toRobot)
            End If

        End If

        'now that we know the other agent's name we can notify the operator
        If Not IsNothing(Me._Operator) Then
            Me._Operator.NotifySignalStrengthReceived(toRobot, pathloss)
        End If

    End Sub

    Protected Overridable Function IsMasterTo(ByVal agentName As String) As Boolean
        Dim idx As Integer = Me._TeamMembers.IndexOf(agentName)
        Return Me._AgentIndex < idx
    End Function

    Protected Function IsSlaveTo(ByVal agentName As String) As Boolean
        Return Not Me.IsMasterTo(agentName)
    End Function

#End Region

#Region " Connection Management "

    Private _ConnectedTo As New Dictionary(Of Guid, String)
    Protected Function IsConnectedTo(ByVal agentName As String) As Boolean
        SyncLock Me._ConnectedTo
            Return Me._ConnectedTo.ContainsValue(agentName)
        End SyncLock
    End Function

    Protected Sub StoreConnectedTo(ByVal conversationID As Guid, ByVal agentName As String)

        Dim notify As Boolean = False
        SyncLock Me._ConnectedTo
            If Not Me._ConnectedTo.ContainsKey(conversationID) Then
                Me._ConnectedTo.Add(conversationID, agentName)
                notify = True
            ElseIf Not Me._ConnectedTo(conversationID) = agentName Then
                Me._ConnectedTo(conversationID) = agentName
                notify = True
            End If
        End SyncLock

        'now that we know the other agent's name we can notify the operator
        If notify AndAlso Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyAgentConnected(agentName)
        End If

    End Sub

    Protected Sub ClearConnectedTo(ByVal conversationID As Guid)
        Dim agentName As String = ""
        SyncLock Me._ConnectedTo
            If Me._ConnectedTo.ContainsKey(conversationID) Then
                agentName = Me._ConnectedTo(conversationID)
                Me._ConnectedTo.Remove(conversationID)
            End If
        End SyncLock

        'notify operator
        If Not String.IsNullOrEmpty(agentName) AndAlso Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyAgentDisconnected(agentName)
        End If

    End Sub

    Protected Function GetConnectedAgentName(ByVal conversationID As Guid) As String
        SyncLock Me._ConnectedTo
            If Me._ConnectedTo.ContainsKey(conversationID) Then
                Return Me._ConnectedTo(conversationID)
            End If
        End SyncLock
        Return String.Empty
    End Function

    Protected Function GetConversation(ByVal agentName As String) As ICommConversation
        Dim conversationID As Guid = Guid.Empty
        SyncLock Me._ConnectedTo
            For Each key As Guid In Me._ConnectedTo.Keys
                If Me._ConnectedTo(key) = agentName Then
                    conversationID = key
                    Exit For
                End If
            Next
        End SyncLock

        Dim conversation As ICommConversation = Nothing
        If Not conversationID = Guid.Empty Then
            SyncLock Me._Conversations
                If Me._Conversations.ContainsKey(conversationID) Then
                    conversation = Me._Conversations(conversationID)
                End If
            End SyncLock
        End If

        Return conversation

    End Function

    Private _Conversations As New Dictionary(Of Guid, ICommConversation)

    Public Sub NotifyConversationStarted(ByVal conversation As ICommConversation) Implements Communication.ICommOwner.NotifyConversationStarted

        SyncLock Me._Conversations
            Me._Conversations.Add(conversation.ConversationID, conversation)
        End SyncLock

        Console.WriteLine(String.Format("[Comm] - {0} sent greetings to [unknown yet]", Me._AgentName))
        conversation.SendMessage(New GreetingMessage(Me._AgentName, Me.Agent.CurrentPoseEstimate))

        'DON'T DO AUTOMATIC SYNCS, THIS BLOWS UP COMM
        'Console.WriteLine(String.Format("[Comm] - {0} sent sync start to [unknown yet]", Me._AgentName))
        'conversation.SendMessage(New SyncStartMessage(Me._AgentName))

    End Sub

    Public Sub NotifyConversationStopped(ByVal conversation As ICommConversation) Implements Communication.ICommOwner.NotifyConversationStopped
        SyncLock Me._Conversations
            Me._Conversations.Remove(conversation.ConversationID)
        End SyncLock
        Me.ClearConnectedTo(conversation.ConversationID)
    End Sub


    Public Sub NotifyMessageReceived(ByVal conversation As ICommConversation, ByVal message As Communication.Message) Implements Communication.ICommOwner.NotifyMessageReceived
        Try
            'check for broadcast messages
            If TypeOf message Is BroadcastMessage Then
                Me.ReceiveBroadcastMessage(conversation, DirectCast(message, BroadcastMessage))
            Else
                Me.ReceiveRegularMessage(conversation, message)
            End If

        Catch ex As Exception
            Console.WriteLine("Error occurred while trying to handle message.")
            Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace)

        End Try
    End Sub

    Protected Overridable Sub ReceiveRegularMessage(ByVal conversation As ICommConversation, ByVal message As Message)
        Select Case message.MessageID

            Case MessageID.Text
                Me.OnTextReceived(conversation, DirectCast(message, TextMessage))

            Case MessageID.Greeting
                Me.OnGreetingReceived(conversation, DirectCast(message, GreetingMessage))
            Case MessageID.ActorCommand
                Me.OnActorCommandReceived(conversation, DirectCast(message, ActorCommandMessage))
            Case MessageID.PoseUpdate
                Me.OnPoseUpdateReceived(conversation, DirectCast(message, PoseUpdateMessage))

            Case MessageID.SyncStart
                Me.OnSyncStartReceived(conversation, DirectCast(message, SyncStartMessage))
            Case MessageID.SyncMaster
                Me.OnSyncMasterReceived(conversation, DirectCast(message, SyncMasterMessage))
            Case MessageID.SyncSlave
                Me.OnSyncSlaveReceived(conversation, DirectCast(message, SyncSlaveMessage))
            Case MessageID.SyncCommit
                Me.OnSyncCommitReceived(conversation, DirectCast(message, SyncCommitMessage))

        End Select
    End Sub

    Protected Overridable Sub OnGreetingReceived(ByVal conversation As ICommConversation, ByVal message As GreetingMessage)
        'store the name of the robot connected to
        Dim otherName As String = message.AgentName
        Me.StoreConnectedTo(conversation.ConversationID, otherName)

        If Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyPoseEstimateReceived(otherName, message.AgentPose)
        End If

    End Sub

    Protected Overridable Sub OnTextReceived(ByVal conversation As ICommConversation, ByVal text As TextMessage)
        'not used atm
    End Sub

#End Region

#Region " Broadcasting Protocol "

    Protected Overridable Sub SendBroadcastMessage(ByVal msg As BroadcastMessage)

        Dim sent As Boolean = False
        If Not msg.ToEveryone Then
            'this message is for a single recipient only

            If Not msg.IsDeliveredTo(msg.Recipient) AndAlso Me.IsConnectedTo(msg.Recipient) Then
                'we are connected to the intended recipient

                Dim conversation As ICommConversation = Me.GetConversation(msg.Recipient)
                If Not IsNothing(conversation) Then
                    Console.WriteLine(String.Format("[Comm] - {0} sent bc-message from {1} to {2}", Me._AgentName, msg.Sender, msg.Recipient))
                    conversation.SendMessage(msg)
                    sent = True
                End If
            End If
        End If

        If Not sent Then

            'either this message is to everyone, or we are not connected to 
            'the intended recipient, or the connection was lost just before
            'we tried to send it, so send to all connected team-members 

            For Each otherName As String In Me._TeamMembers
                If Not msg.IsDeliveredTo(otherName) AndAlso Me.IsConnectedTo(otherName) Then
                    Dim conv As ICommConversation = Me.GetConversation(otherName)
                    If Not IsNothing(conv) Then
                        Console.WriteLine(String.Format("[Comm] - {0} sent bc-message from {1} to {2}", Me._AgentName, msg.Sender, otherName))
                        conv.SendMessage(msg)
                    End If
                End If
            Next

        End If
    End Sub

    Protected Overridable Sub ReceiveBroadcastMessage(ByVal conversation As ICommConversation, ByVal msg As BroadcastMessage)

        'mark this message as delivered to me, to avoid resends
        msg.MarkDelivered(Me._AgentName)

        If msg.ToEveryone OrElse msg.Recipient = Me._AgentName Then
            'this message (also) is for me! forward to regular message handling
            Console.WriteLine(String.Format("[Comm] - bc-message from {0} to {1} reached destination: {2}", msg.Sender, msg.Recipient, Me._AgentName))
            Me.ReceiveRegularMessage(conversation, msg)
        End If

        If msg.ToEveryone OrElse (Not msg.IsDeliveredTo(msg.Recipient) AndAlso Not msg.Recipient = Me._AgentName) Then
            'this message needs further proceccessing

            Dim sent As Boolean = False
            If Not msg.ToEveryone AndAlso Me.IsConnectedTo(msg.Recipient) Then
                'nice! we are connected to the intended recipient
                Dim conv As ICommConversation = Me.GetConversation(msg.Recipient)
                If Not IsNothing(conv) Then
                    Console.WriteLine(String.Format("[Comm] - {0} forwarded bc-message from {1} to {2}", Me._AgentName, msg.Sender, msg.Recipient))
                    conv.SendMessage(msg)
                    sent = True
                End If
            End If

            If Not sent Then

                'either this message is to everyone, or we are not connected to 
                'the intended recipient, or the connection was lost just before
                'we tried to send it, so forward to all connected team-members 

                For Each otherName As String In Me._TeamMembers
                    If Not msg.IsDeliveredTo(otherName) AndAlso Me.IsConnectedTo(otherName) Then
                        Dim conv As ICommConversation = Me.GetConversation(otherName)
                        If Not IsNothing(conv) Then
                            Console.WriteLine(String.Format("[Comm] - {0} forwarded bc-message from {1} to {2}", Me._AgentName, msg.Sender, otherName))
                            conv.SendMessage(msg)
                        End If
                    End If
                Next

            End If
        End If
    End Sub

#End Region

#Region " Relay Actor Commands and Pose Updates "

    Public Sub RelayActorCommand(ByVal toRobot As String, ByVal command As String)
        Me.SendBroadcastMessage(New ActorCommandMessage(Me._AgentName, toRobot, command))
    End Sub

    Protected Overridable Sub OnActorCommandReceived(ByVal conversation As ICommConversation, ByVal message As ActorCommandMessage)
        'forward to agent
        If Not IsNothing(Me.Agent) Then
            Console.WriteLine(String.Format("[Comm] -- {0} alter ego received actor command '{1}'", Me._AgentName, message.Command))
            Me.Agent.SendUsarSimCommand(message.Command)
        End If
    End Sub


    Public Sub RelayPoseEstimateUpdate(ByVal pose As Math.Pose2D)
        Me.SendBroadcastMessage(New PoseUpdateMessage(Me._AgentName, Me._OperatorName, pose))
    End Sub

    Protected Overridable Sub OnPoseUpdateReceived(ByVal conversation As ICommConversation, ByVal message As PoseUpdateMessage)

        'store the name of the robot connected to
        'Dim otherName As String = message.AgentName
        Dim otherName As String = Me.GetConnectedAgentName(conversation.ConversationID)
        Dim robotName As String = message.Sender
        Console.WriteLine(String.Format("[Comm] - {0} received pose update for {1} from {2}", Me._AgentName, robotName, otherName))

        If Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyPoseEstimateReceived(robotName, message.Pose)
        End If

    End Sub

#End Region

#Region " Manifold Synchronization "

    Public Sub RelaySyncRequest(ByVal toRobot As String)
        Dim conversation As ICommConversation = Me.GetConversation(toRobot)
        If Not IsNothing(conversation) Then
            conversation.SendMessage(New SyncStartMessage(Me.Agent.CurrentPoseEstimate))
        End If
    End Sub



    Private _LastConfirmedUpdate As New Dictionary(Of String, DateTime)

    Protected Overridable Sub OnSyncStartReceived(ByVal conversation As ICommConversation, ByVal message As SyncStartMessage)

        Dim otherName As String = Me.GetConnectedAgentName(conversation.ConversationID)
        Console.WriteLine(String.Format("[Comm] - {0} received sync start from {1}", Me._AgentName, otherName))

        If Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyPoseEstimateReceived(otherName, message.CurrentPose)
        End If

        'first the slave will sync the master
        If Me.IsSlaveTo(otherName) Then

            'get the timestamp of the last confirmed update 
            '(retrieved by a SyncCommit message)
            Dim lastUpdate As DateTime = DateTime.MinValue
            SyncLock Me._LastConfirmedUpdate
                If Me._LastConfirmedUpdate.ContainsKey(otherName) Then
                    lastUpdate = Me._LastConfirmedUpdate(otherName)
                End If
            End SyncLock

            'collect changes since last update
            Me._Manifold.AcquireReaderLock()
            Dim response As New SyncMasterMessage
            If Not Me.LoadDataForSyncDataMessage(lastUpdate, response) Then
                response = Nothing
            End If
            Me._Manifold.ReleaseReaderLock()

            'submit update
            If Not IsNothing(response) Then
                Console.WriteLine(String.Format("[Comm] - {0} sent sync update to {1} ({2} items since {3})", Me._AgentName, otherName, response.Data.Count, lastUpdate))
                conversation.SendMessage(response)
            End If

        End If

    End Sub

    Protected Overridable Sub OnSyncMasterReceived(ByVal conversation As ICommConversation, ByVal message As SyncMasterMessage)

        Dim otherName As String = Me.GetConnectedAgentName(conversation.ConversationID)

        'get the timestamp of the last confirmed update 
        '(retrieved by a SyncCommit message)
        Dim lastUpdate As DateTime = DateTime.MinValue
        SyncLock Me._LastConfirmedUpdate
            If Me._LastConfirmedUpdate.ContainsKey(otherName) Then
                lastUpdate = Me._LastConfirmedUpdate(otherName)
            End If
        End SyncLock

        If lastUpdate = DateTime.MinValue Then
            Console.WriteLine(String.Format("[Comm] - {0} received first sync update from {1} at {2})", Me._AgentName, otherName, DateTime.Now))
        End If

        Me._Manifold.AcquireWriterLock()

        'before applying the sync update, create the response for the slave
        Dim response As New SyncSlaveMessage
        If Not Me.LoadDataForSyncDataMessage(lastUpdate, response) Then
            response = Nothing
        End If

        'within the same lock, copy all received items into the manifold
        Console.WriteLine(String.Format("[Comm] - {0} received sync update from {1} ({2} items)", Me._AgentName, otherName, message.Data.Count))
        Dim memento As IMemento
        For Each item As Object In message.Data
            If TypeOf item Is IMemento Then
                memento = DirectCast(item, IMemento)
                memento.Restore(Me._Manifold)
            End If
        Next

        'get the timestamp that the slave must respond with upon commit
        If Not IsNothing(response) Then
            response.CommitTimestamp = DateTime.Now
        End If

        Me._Manifold.ReleaseWriterLock()

        'commit the changes to slave, so the slave knows what to send next time
        Console.WriteLine(String.Format("[Comm] - {0} sent sync commit to {1} ({2})", Me._AgentName, otherName, message.CommitTimestamp))
        conversation.SendMessage(New SyncCommitMessage(message.CommitTimestamp, Me.Agent.CurrentPoseEstimate))

        'and now send sync-update to slave
        If Not IsNothing(response) Then
            Console.WriteLine(String.Format("[Comm] - {0} sent sync update to {1} ({2} items since {3})", Me._AgentName, otherName, response.Data.Count, lastUpdate))
            conversation.SendMessage(response)
        End If

    End Sub

    Protected Overridable Sub OnSyncSlaveReceived(ByVal conversation As ICommConversation, ByVal message As SyncSlaveMessage)

        Dim otherName As String = Me.GetConnectedAgentName(conversation.ConversationID)

        Me._Manifold.AcquireWriterLock()

        'copy all received items into the manifold
        Console.WriteLine(String.Format("[Comm] - {0} received sync update from {1} ({2} items)", Me._AgentName, otherName, message.Data.Count))
        Dim memento As IMemento
        For Each item As Object In message.Data
            If TypeOf item Is IMemento Then
                memento = DirectCast(item, IMemento)
                memento.Restore(Me._Manifold)
            End If
        Next

        'this data was received from master, and the sync protocol 
        'guarantees that the master was updated already, to avoid sending
        'back the recent changes in the next sync, update the timestamp for master
        Dim timestamp As DateTime = Now
        SyncLock Me._LastConfirmedUpdate
            If Me._LastConfirmedUpdate.ContainsKey(otherName) Then
                Me._LastConfirmedUpdate(otherName) = timestamp
            Else
                Me._LastConfirmedUpdate.Add(otherName, timestamp)
            End If
        End SyncLock

        Me._Manifold.ReleaseWriterLock()

        'now commit the recent changes to master, so that master knows what to send next time
        Console.WriteLine(String.Format("[Comm] - {0} sent sync commit to {1} ({2})", Me._AgentName, otherName, message.CommitTimestamp))
        conversation.SendMessage(New SyncCommitMessage(message.CommitTimestamp, Me.Agent.CurrentPoseEstimate))

    End Sub

    Protected Overridable Sub OnSyncCommitReceived(ByVal conversation As ICommConversation, ByVal message As SyncCommitMessage)

        Dim otherName As String = Me.GetConnectedAgentName(conversation.ConversationID)
        Dim timestamp As DateTime = message.Timestamp

        Console.WriteLine(String.Format("[Comm] - {0} received sync commit from {1} ({2})", Me._AgentName, otherName, timestamp))

        If Not IsNothing(Me._Operator) Then
            Me._Operator.NotifyPoseEstimateReceived(otherName, message.CurrentPose)
        End If

        'store timestamp 
        SyncLock Me._LastConfirmedUpdate
            If Me._LastConfirmedUpdate.ContainsKey(otherName) Then
                Me._LastConfirmedUpdate(otherName) = timestamp
            Else
                Me._LastConfirmedUpdate.Add(otherName, timestamp)
            End If
        End SyncLock

    End Sub



    '    Protected Overridable Sub OnSyncRequestReceived(ByVal conversation As ICommConversation, ByVal message As SyncRequestMessage)

    '        'construct response message with updates since timestamp
    '        Dim msg As SyncResponseMessage = Me.CreateResponseMessage(message.Timestamp)

    '        'the master will cache the response for now, and send it later when 
    '        'the slave communicates that it's ready. What it does now instead is send
    '        'the previously cached request.
    '        'the slave will immediately send the response back to master.

    '        Dim name As String = Me.GetConnectedAgentName(conversation.ConversationID)
    '        If Me.IsMasterTo(name) Then

    '            'the master keeps the response in a cache
    '            Console.WriteLine(String.Format("[Comm] - {0} buffered response for {1}", Me._AgentName, name))
    '            SyncLock Me._Responses
    '                Me._Responses(conversation.ConversationID) = msg
    '            End SyncLock

    '            'and the previously cached request is sent instead
    '            Dim request As SyncRequestMessage = Nothing
    '            SyncLock Me._Requests
    '                If Me._Requests.ContainsKey(conversation.ConversationID) Then
    '                    request = Me._Requests(conversation.ConversationID)
    '                End If
    '            End SyncLock

    '            If Not IsNothing(request) Then
    '                Console.WriteLine(String.Format("[Comm] - {0} sent buffered request to {1} ({2})", Me._AgentName, name, request.Timestamp))
    '                conversation.SendMessage(request)
    '            End If

    '        Else
    '            'the slave will simply reply to the request by sending the response
    '            Console.WriteLine(String.Format("[Comm] - {0} sent response to {1} ({2})", Me._AgentName, name, msg.Data.Count))
    '            conversation.SendMessage(msg)

    '        End If

    '    End Sub

    '    Protected Overridable Sub OnSyncResponseReceived(ByVal conversation As ICommConversation, ByVal message As SyncResponseMessage)

    '        'to keep track of the most recent update we recieved
    '        'default to Now, so that we will at minimum communicate
    '        'the current timestamp back to the sender.
    '        Dim timestamp As DateTime = DateTime.MinValue

    '        'copy all received items into the manifold
    '        Dim memento As IMemento
    '        For Each item As Object In message.Data
    '            If TypeOf item Is IMemento Then

    '                memento = DirectCast(item, IMemento)
    '                memento.Restore(Me._Manifold)

    '                'keep track of most recent timestamp
    '                If memento.Timestamp > timestamp Then
    '                    timestamp = memento.Timestamp
    '                End If

    '            End If
    '        Next

    '        Dim name As String = Me.GetConnectedAgentName(conversation.ConversationID)

    '        'tell that we are ready processing the response
    '        Console.WriteLine(String.Format("[Comm] - {0} sent ready to {1}", Me._AgentName, name))
    '        conversation.SendMessage(New SyncReadyMessage(timestamp))

    '        'the master will now send back his previously cached response

    '        If Me.IsMasterTo(name) Then

    '            'we still had the response for slave in cache, now we will send it
    '            Dim response As SyncResponseMessage = Nothing
    '            SyncLock Me._Responses
    '                If Me._Responses.ContainsKey(conversation.ConversationID) Then
    '                    response = Me._Responses(conversation.ConversationID)
    '                End If
    '            End SyncLock

    '            If Not IsNothing(response) Then
    '                Console.WriteLine(String.Format("[Comm] - {0} sent buffered response to {1} ({2})", Me._AgentName, name, response.Data.Count))
    '                conversation.SendMessage(response)
    '            End If

    '        End If

    '    End Sub

    '    Protected Overridable Sub OnSyncReadyReceived(ByVal conversation As ICommConversation, ByVal message As SyncReadyMessage)

    '        Dim agentName As String = Me.GetConnectedAgentName(conversation.ConversationID)
    '        Console.WriteLine(String.Format("[Comm] - {0} retrieved ready from {1}", Me._AgentName, agentName))

    '        Dim timestamp As DateTime = message.Timestamp
    '        If timestamp > DateTime.MinValue Then

    '            'store timestamp 
    '            Console.WriteLine(String.Format("[Comm] - {0} stored timestamp for {1}: {2}", Me._AgentName, agentName, timestamp))

    '            'the next time we send a request we will only request for updates more recent
    '            'then this timestamp
    '            SyncLock Me._LastConfirmedUpdate
    '                If Me._LastConfirmedUpdate.ContainsKey(agentName) Then
    '                    Me._LastConfirmedUpdate(agentName) = timestamp
    '                Else
    '                    Me._LastConfirmedUpdate.Add(agentName, timestamp)
    '                End If
    '            End SyncLock

    '        End If


    '        'the master will wait for 5 secs and then trigger another sync-cycle
    '        If Me.IsMasterTo(agentName) Then

    '            'pause for a while
    '            Console.WriteLine(String.Format("--- [Comm] - {0} starts a new sync cycle with {1} in {2} secs", Me._AgentName, agentName, RESYNC_TIMEOUT))

    '            Thread.Sleep(TimeSpan.FromSeconds(RESYNC_TIMEOUT))

    '            'trigger another sync cycle
    '            Console.WriteLine(String.Format("[Comm] - {0} sent greetings to {1}", Me._AgentName, agentName))
    '            conversation.SendMessage(New GreetingMessage(Me._AgentName, Me.Agent.CurrentPoseEstimate))

    '        End If

    '    End Sub

#End Region

#Region " Helper Functions "

    ''' <summary>
    ''' Helper function that will construct a message with all changes
    ''' since the timestamp.
    ''' </summary>
    ''' <param name="lastUpdate"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Protected Function LoadDataForSyncDataMessage(ByVal lastUpdate As DateTime, ByVal msg As SyncDataMessage) As Boolean

        Try

            'patches must go first, otherwise the relations cannot restore themselves

            For Each patch As Patch In Me._Manifold.Nodes
                If patch.Timestamp > lastUpdate Then
                    msg.Data.Add(patch.CreateMemento)
                End If
            Next

            For Each relation As Relation In Me._Manifold.Links
                If relation.Timestamp > lastUpdate Then
                    msg.Data.Add(relation.CreateMemento)
                End If
            Next

            msg.CommitTimestamp = DateTime.Now

            Return True

        Catch ex As Exception
            Console.WriteLine(ex)
        End Try

        Return False

    End Function

#End Region

End Class
