Imports System.IO
Imports System.Drawing
Imports System.Threading

Imports UvARescue.Math
Imports UvARescue.Tools

''' <summary>
''' Extends graph with locking capabilities and specific Manifold-related 
''' functionality.
''' 
''' Please note that although locking functionality is provided, it is NOT
''' used in the Manifold's members. So the Manifold (like the Graph from which
''' it inherits) is NOT thread-safe UNLESS this is enforced by all consuming 
''' parties.
''' </summary>
''' <remarks></remarks>
Public Class Manifold
    Inherits Graph(Of Patch, Relation)

#Region " Observers "

    Private observers As New List(Of IManifoldObserver)
    Public Sub AddObserver(ByVal observer As IManifoldObserver)
        SyncLock Me.observers
            Me.observers.Add(observer)
        End SyncLock
    End Sub
    Public Sub RemoveObserver(ByVal observer As IManifoldObserver)
        SyncLock Me.observers
            Me.observers.Remove(observer)
        End SyncLock
    End Sub

    Protected Overrides Sub OnNodeInserted(ByVal node As Patch)
        MyBase.OnNodeInserted(node)
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyPatchInserted(node)
            Next
        End SyncLock
    End Sub

    Protected Overrides Sub OnLinkInserted(ByVal link As Relation)
        MyBase.OnLinkInserted(link)
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyRelationInserted(link)
            Next
        End SyncLock

        Dim nodes As Integer
        Dim links As Integer

        Me.AcquireReaderLock()
        nodes = Me.NodeCount
        links = Me.LinkCount
        Me.ReleaseReaderLock()

        Console.WriteLine(String.Format("[Manifold]-- Extended, currently has  {0} patches / {1} relations", nodes, links))

    End Sub

    Public Overrides Sub Clear()
        MyBase.Clear()
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyCleared()
            Next
        End SyncLock
    End Sub

#End Region

#Region " Locking "

    'reader-writer locks allow either a single writer (x)or multiple simultaneous
    'readers. Since heavy (read-only) use of the manifold is anticipated
    '(especially for Gui rendering) this made reader-writer locks the preferred
    'choice over mutexes or other mutual exclusive synchronization instruments 
    'like wait-handles.


    Private lock As New ReaderWriterLock

    Public Sub AcquireReaderLock()
        Me.lock.AcquireReaderLock(Timeout.Infinite)
        'Console.WriteLine(Thread.CurrentThread.Name & " acquired Read lock")
    End Sub
    Public Sub ReleaseReaderLock()
        'Console.WriteLine(Thread.CurrentThread.Name & " released Read lock")
        Me.lock.ReleaseReaderLock()
    End Sub

    Public Sub AcquireWriterLock()
        Me.lock.AcquireWriterLock(Timeout.Infinite)
        'Console.WriteLine(Thread.CurrentThread.Name & " acquired Write lock")
    End Sub
    Public Sub ReleaseWriterLock()
        'Console.WriteLine(Thread.CurrentThread.Name & " released Write lock")
        Me.lock.ReleaseWriterLock()
    End Sub

    Public Sub ReleaseLock()
        Me.lock.ReleaseLock()
    End Sub

#End Region

#Region " Mapping Functionality "

    Public Overridable Sub Extend(ByVal newPatch As Patch)
        Me.InsertNode(newPatch)
    End Sub

    Public Overridable Sub Extend(ByVal atPatch As Patch, ByVal newPatch As Patch, ByVal odometry As Pose2D, ByVal covariance As Matrix3)
        Me.Extend(newPatch, New Relation(Me, atPatch, newPatch, odometry, covariance))
    End Sub

    Public Overridable Sub Extend(ByVal newPatch As Patch, ByVal newRelation As Relation)
        Me.InsertNode(newPatch)
        Me.InsertLink(newRelation)
    End Sub

    Public Overridable Sub Extend(ByVal newRelation As Relation)
        Me.InsertLink(newRelation)
    End Sub

#End Region

#Region " Localization Functionality "

    Private TrackedAgents As New Dictionary(Of Guid, Agent)
    Private CurrentPatches As New Dictionary(Of Guid, Guid)
    Private CurrentPoses As New Dictionary(Of Guid, Pose2D)

    Public Function GetTrackedAgents() As IEnumerable(Of Agent)
        Return Me.TrackedAgents.Values
    End Function

    Public Function GetCurrentPatch(ByVal agent As Agent) As Patch
        If Not Me.CurrentPatches.ContainsKey(agent.UniqueID) Then
            Return Nothing
        End If
        Return Me.GetNode(Me.CurrentPatches(agent.UniqueID))
    End Function
    Public Function GetCurrentPose(ByVal agent As Agent) As Pose2D
        If Not Me.CurrentPoses.ContainsKey(agent.UniqueID) Then
            Return Nothing
        End If
        Return Me.CurrentPoses(agent.UniqueID)
    End Function

    Public Overridable Sub LocalizeAgent(ByVal agent As Agent, ByVal patch As Patch, ByVal pose As Pose2D)

        If Not Me.TrackedAgents.ContainsKey(agent.UniqueID) Then
            Me.TrackedAgents.Add(agent.UniqueID, agent)
        End If

        Me.CurrentPoses(agent.UniqueID) = pose
        If Not IsNothing(patch) Then
            Me.CurrentPatches(agent.UniqueID) = patch.ID
        Else
            Me.CurrentPatches(agent.UniqueID) = Guid.Empty
        End If

        agent.NotifyPoseEstimateUpdate(pose)

        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyAgentMoved(agent, pose)
            Next
        End SyncLock

    End Sub

#End Region

#Region " Victim Observations "

    Private _Victims As New Dictionary(Of String, VictimObservation)
    Public ReadOnly Property Victims() As IEnumerable(Of VictimObservation)
        Get
            Return Me._Victims.Values
        End Get
    End Property
    Public ReadOnly Property VictimCount() As Integer
        Get
            Return Me._Victims.Count
        End Get
    End Property

    Public Function FindNearestVictim(ByVal x As Double, ByVal y As Double, ByVal maxDistance As Double) As VictimObservation

        Dim curDistance As Double = Double.MaxValue
        Dim curVictim As VictimObservation = Nothing

        For Each victim As VictimObservation In Me._Victims.Values
            Dim distance As Double = ((x - victim.AverageX) ^ 2 + (y - victim.AverageY) ^ 2) ^ 0.5
            If distance < curDistance Then
                curDistance = distance
                curVictim = victim
            End If
        Next

        'check if this observation is within the specified maximum distance
        If curDistance <= maxDistance Then
            Return curVictim
        Else
            Return Nothing
        End If

    End Function

    Public Function GetVictim(ByVal victimID As String) As VictimObservation
        If Me._Victims.ContainsKey(victimID) Then
            Return Me._Victims(victimID)
        End If
        Return Nothing
    End Function

    Public Function HasVictim(ByVal victimID As String) As Boolean
        Return Me._Victims.ContainsKey(victimID)
    End Function

    Public Function InserVictim(ByVal victim As VictimObservation) As VictimObservation
        Me._Victims.Add(victim.VictimID, victim)
        Me.OnVictimInserted(victim)
        Return victim
    End Function

    Public Function RemoveVictim(ByVal victimID As String) As VictimObservation
        Dim victim As VictimObservation = Me.GetVictim(victimID)
        If Not IsNothing(victim) Then
            Me._Victims.Remove(victimID)
            Me.OnVictimRemoved(victim)
        End If
        Return victim
    End Function


    ''' <summary>
    ''' Invoked by VictimObservations to notify the manifold of updates.
    ''' </summary>
    ''' <param name="victim"></param>
    ''' <remarks></remarks>
    Friend Sub NotifyVictimUpdated(ByVal victim As VictimObservation)
        Me.OnVictimUpdated(victim)
    End Sub

    Protected Overridable Sub OnVictimInserted(ByVal victim As VictimObservation)
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyVictimUpdated(victim)
            Next
        End SyncLock
    End Sub
    Protected Overridable Sub OnVictimUpdated(ByVal victim As VictimObservation)
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyVictimUpdated(victim)
            Next
        End SyncLock
    End Sub
    Protected Overridable Sub OnVictimRemoved(ByVal victim As VictimObservation)
        SyncLock Me.observers
            For Each observer As IManifoldObserver In Me.observers
                observer.NotifyVictimUpdated(victim)
            Next
        End SyncLock
    End Sub

#End Region

#Region " WriteVictFile "

    Public Sub WriteVictFile()

        ' define filename and check existence directory
        Dim pathName, victFileName, imFileName, logFileName As String
        pathName = Path.Combine(My.Application.Info.DirectoryPath, "Logs\")
        victFileName = Path.Combine(pathName, "Victim.txt")
        Dim fileInfo As New FileInfo(pathName)
        If Not fileInfo.Directory.Exists Then
            fileInfo.Directory.Create()
        End If

        ' write victims to files
        Using sw As StreamWriter = New StreamWriter(victFileName)
            For Each victim As String In Me._Victims.Keys
                sw.Write(String.Format("{0}, ", victim))
                sw.Write(String.Format("{0:f2}, ", Me._Victims(victim).AverageX / 1000))
                sw.Write(String.Format("{0:f2}, ", Me._Victims(victim).AverageY / 1000))
                sw.Write(String.Format("{0}, ", 0))
                sw.WriteLine(String.Format("{0}", Me._Victims(victim).Method))

                Dim pic As Bitmap = Me._Victims(victim).Picture
                If Not pic Is Nothing Then
                    imFileName = Path.Combine(pathName, String.Format("{0}.jpg", victim))
                    pic.Save(imFileName, System.Drawing.Imaging.ImageFormat.Jpeg)
                End If
            Next
            sw.WriteLine("END")
        End Using

        logFileName = Path.Combine(pathName, "VictLogFile.txt")
        Dim i As Integer
        Using sw As StreamWriter = New StreamWriter(logFileName)
            For Each victim As String In Me._Victims.Keys
                i = 0
                sw.WriteLine(String.Format("[{0}] Log Entries:", victim))
                While i < Me._Victims(victim).LogEntries.Count
                    sw.WriteLine(Me._Victims(victim).LogEntries(i))
                    i += 1
                End While

            Next
        End Using
    End Sub

#End Region

End Class
