Imports System.Drawing
Imports System.Math

Imports UvARescue.Agent
Imports UvARescue.Math
Imports UvARescue.Tools

''' <summary>
''' A patch stores a single laser range scan with a pose estimate.
''' </summary>
''' <remarks></remarks>
Public Class Patch
    Inherits GraphNode(Of Patch, Relation)

#Region " Constructor "

    Public Sub New(ByVal manifold As Manifold, ByVal scan As ScanObservation, ByVal rawOrigin As Pose2D)
        MyBase.New(manifold, Guid.NewGuid)
        Me._Timestamp = Now
        Me._Scan = scan
        Me._RawOrigin = rawOrigin
        Me._EstimatedOrigin = rawOrigin
    End Sub

    Protected Sub New(ByVal manifold As Manifold, ByVal patchID As Guid, ByVal scan As ScanObservation, ByVal rawOrigin As Pose2D)
        MyBase.New(manifold, patchID)
        Me._Timestamp = Now
        Me._Scan = scan
        Me._RawOrigin = rawOrigin
        Me._EstimatedOrigin = rawOrigin
    End Sub


#End Region

#Region " Memento "

    <Serializable()> _
    Protected Class PatchMemento
        Implements IMemento

        Public Sub New(ByVal patch As Patch)
            Me._PatchID = Patch.ID
            Me._Scan = Patch._Scan
            Me._RawOrigin = Patch._RawOrigin
            Me._EstimatedOrigin = Patch._EstimatedOrigin

            'copy all victim data ever processed
            Me._VictimPoses.AddRange(Patch._VictimPoses)
            Me._VictimData.AddRange(Patch._VictimData)
            Me._VictimTimestamps.AddRange(Patch._VictimTimestamps)

            Me._Timestamp = Patch.Timestamp

            Me._PicturePoses.AddRange(Patch._PicturePoses)
            Me._PictureX.AddRange(Patch._PictureXs)
            Me._PictureY.AddRange(Patch._PictureYs)
            Me._HistIms.AddRange(Patch._HistIms)
            Me._SkinPercentage.AddRange(Patch._SkinPercentages)
            Me._ImageTimestamps.AddRange(Patch._VictimTimestamps)

        End Sub

        Private _PatchID As Guid
        Private _Scan As ScanObservation
        Private _RawOrigin As Pose2D
        Private _EstimatedOrigin As Pose2D

        Private _VictimPoses As New List(Of Pose2D)
        Private _VictimData As New List(Of VictimData)
        Private _VictimTimestamps As New List(Of DateTime)

        Private _PicturePoses As New List(Of Pose2D)
        Private _PictureX As New List(Of Double)
        Private _PictureY As New List(Of Double)
        Private _HistIms As New List(Of Bitmap)
        Private _SkinPercentage As New List(Of Double)
        Private _ImageTimestamps As New List(Of DateTime)


        Private _Timestamp As DateTime
        Public ReadOnly Property Timestamp() As Date Implements IMemento.Timestamp
            Get
                Return Me._Timestamp
            End Get
        End Property


        Public Sub Restore(ByVal manifold As Manifold) Implements IMemento.Restore
            If Not manifold.HasNode(Me._PatchID) Then
                Dim patch As New Patch(manifold, Me._PatchID, Me._Scan, Me._RawOrigin)
                patch._EstimatedOrigin = Me._EstimatedOrigin
                manifold.Extend(patch)

                're-process all victim data 
                Dim i As Integer = 0
                While i < Me._VictimData.Count
                    patch.ProcessVictimData(Me._VictimPoses(i), Me._VictimData(i))
                    i += 1
                End While

                i = 0
                While i < Me._PicturePoses.Count
                    patch.ProcessVictimPicture(Me._PicturePoses(i), Me._PictureX(i), Me._PictureY(i), Me._HistIms(i), Me._SkinPercentage(i))
                    i += 1
                End While

                'IMPORTANT
                'do not update timestamp, every manifold must assign its own
                'timestamps to patches in order to make manifold sharing
                'between more than 2 robots work!
                'patch._Timestamp = Me._Timestamp


            End If
        End Sub

    End Class

    Public Function CreateMemento() As IMemento
        Return New PatchMemento(Me)
    End Function

#End Region

#Region " Patch Data "

    Public ReadOnly Property Manifold() As Manifold
        Get
            Return DirectCast(Me.Graph, Manifold)
        End Get
    End Property

    Private _Timestamp As DateTime
    Public ReadOnly Property Timestamp() As DateTime
        Get
            Return Me._Timestamp
        End Get
    End Property

    Private _Scan As ScanObservation
    Public ReadOnly Property Scan() As ScanObservation
        Get
            Return _Scan
        End Get
    End Property

    Private _RawOrigin As Pose2D
    Public ReadOnly Property RawOrigin() As Pose2D
        Get
            Return _RawOrigin
        End Get
    End Property

    Private _EstimatedOrigin As Pose2D

    ''' <summary>
    ''' This property is publicly writable, however it is typically solely
    ''' up to a SLAM implementation to maintain the estimate.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property EstimatedOrigin() As Pose2D
        Get
            Return _EstimatedOrigin
        End Get
        Set(ByVal value As Pose2D)
            Me._EstimatedOrigin = value
        End Set
    End Property

#End Region

#Region " Cached Properties "

    Private _Points() As Vector2 = Nothing
    Public ReadOnly Property Points() As Vector2()
        Get

            'since points are in local coords they are independent of the current
            'pose estimate. Hence, once computed they can be cached.

            If IsNothing(Me._Points) Then
                'compute the points for the first time.
                With Me.Scan
                    Dim angles() As Double = Me.Angles
                    ReDim Me._Points(.Length - 1)

                    Dim dist As Double
                    Dim angle As Double
                    For i As Integer = 0 To .Length - 1
                        dist = .Range(i)
                        angle = angles(i)
                        Me._Points(i) = New Vector2(Cos(angle) * (dist * .Factor), Sin(angle) * (dist * .Factor))
                    Next

                End With
            End If
            Return Me._Points
        End Get
    End Property

    Private _Angles() As Double = Nothing
    Public ReadOnly Property Angles() As Double()
        Get
            'cachable, just like Points
            If IsNothing(Me._Angles) Then
                'compute the angles
                With Me.Scan
                    Dim fromAngle As Double = -.FieldOfView / 2
                    ReDim Me._Angles(.Length - 1)
                    For i As Integer = 0 To .Length - 1
                        Me._Angles(i) = fromAngle + .Resolution * i
                    Next
                End With
            End If
            Return Me._Angles
        End Get
    End Property

#End Region

#Region " Projections "

    ''' <summary>
    ''' Converts the Points to global coords based on the current estimate
    ''' of the origin.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function ScanToGlobal() As Vector2()

        Dim rotmx As TMatrix2D = Me.EstimatedOrigin.ToGlobalMatrix

        Dim points() As Vector2 = Me.Points
        Dim pglobal(points.Length - 1) As Vector2
        For i As Integer = 0 To points.Length - 1
            pglobal(i) = rotmx * points(i)
        Next
        Return pglobal

    End Function

#End Region

#Region " Victim Observations "

    'to keep track of all the victim-related data that we have seen
    Private _VictimPoses As New List(Of Pose2D)
    Private _VictimData As New List(Of VictimData)
    Private _VictimTimestamps As New List(Of DateTime)

    Private _VictimParts As New Dictionary(Of String, List(Of VictimData.VictimPart))

    Private _histIm As Bitmap

    Public Sub ProcessVictimData(ByVal pose As Pose2D, ByVal victimData As VictimData)

        Me._VictimPoses.Add(pose)
        Me._VictimData.Add(victimData)
        Me._VictimTimestamps.Add(Now)

        Dim factor As Single = 1000 'to convert from m to mm

        For Each part As VictimData.VictimPart In victimData.Parts

            If part.PartName = "chest" Or part.PartName = "head" Or part.PartName = "leg" Or part.PartName = "hand" Then

                'note that the part is observed in local coords wrt the current pose of the robot
                'transform to global coordinate wrt pose using a Pose2D object with dummy rotation
                Dim pglobal As Vector2 = New Pose2D(CType(part.X * factor, Single), CType(part.Y * factor, Single), 0).ToGlobal(pose).Position

                'find observation nearest to this global coord
                Dim observation As VictimObservation = Me.Manifold.FindNearestVictim(pglobal.X, pglobal.Y, 2000)
                Dim victimID As String

                If Not IsNothing(observation) Then
                    'a nearby observation was found, contribute to it
                    victimID = observation.VictimID
                Else
                    'no nearby observation was found, create a new one
                    victimID = String.Format("VictimZone-{0}", Me.Manifold.VictimCount)
                    observation = Me.Manifold.InserVictim(New VictimObservation(Me.Manifold, victimID, "VSENSOR"))
                End If

                Dim count As Integer = victimData.PartCount(part.PartName)

                observation.AcquireWriterLock()
                observation.AddVictimPart(part, count, pglobal.X, pglobal.Y)
                observation.AddLogEntry(String.Format("[{0}]-- Agent at {1}: seen part {2} of {3} at location {4:F2},{5:F2}", Now.ToString, Me.RawOrigin.ToString, part.PartName, victimID, pglobal.X / 1000, pglobal.Y / 1000))
                observation.ReleaseWriterLock()

                'also store relevant information in local dictionary

                Dim list As List(Of VictimData.VictimPart)
                If Me._VictimParts.ContainsKey(victimID) Then
                    list = Me._VictimParts(victimID)
                Else
                    list = New List(Of VictimData.VictimPart)
                    Me._VictimParts(victimID) = list
                End If

                list.Add(part)

            End If

        Next

        'update timestamp
        Me._Timestamp = Now

    End Sub

    Private _PicturePoses As New List(Of Pose2D)
    Private _PictureXs As New List(Of Double)
    Private _PictureYs As New List(Of Double)
    Private _HistIms As New List(Of Bitmap)
    Private _SkinPercentages As New List(Of Double)

    Public Sub ProcessVictimPicture(ByVal pose As Pose2D, ByVal x As Double, ByVal y As Double, ByVal histIm As Bitmap, ByVal skinPercentage As Double)

        Me._PicturePoses.Add(pose)
        Me._PictureXs.Add(x)
        Me._PictureYs.Add(y)
        Me._HistIms.Add(histIm)
        Me._SkinPercentages.Add(skinPercentage)

        Me._VictimTimestamps.Add(Now)

        Dim factor As Single = 1000 'to convert from m to mm

        'note that the part is observed in local coords wrt the current pose of the robot
        'transform to global coordinate wrt pose using a Pose2D object with dummy rotation
        Dim pglobal As Vector2 = New Pose2D(CType(x * factor, Single), CType(y * factor, Single), 0).ToGlobal(pose).Position

        'find observation nearest to this global coord
        Dim observation As VictimObservation = Me.Manifold.FindNearestVictim(pglobal.X, pglobal.Y, 2000)
        Dim victimID As String

        If Not IsNothing(observation) Then
            'a nearby observation was found, contribute to it
            victimID = observation.VictimID
        Else
            'no nearby observation was found, create a new one
            victimID = String.Format("VictimZone-{0}", Me.Manifold.VictimCount)
            observation = Me.Manifold.InserVictim(New VictimObservation(Me.Manifold, victimID, "VISUAL"))
        End If

        observation.AcquireWriterLock()
        If observation.UpdatePicture(histIm, skinPercentage) Then
            'add "visual" body part
            Dim avgLoc As String = String.Format("{0},{1},0", x, y)
            Dim part As VictimData.VictimPart = New VictimData.VictimPart("Visual", avgLoc)
            observation.AddVictimPart(part, 1, pglobal.X, pglobal.Y)
            observation.AddLogEntry(String.Format("[{0}]-- Agent at {1}: Seen image of {2} having {3:F2} amount of skin", Now.ToString, Me.RawOrigin.ToString, victimID, skinPercentage))
            'add histIm to patch
            Me._histIm = histIm
            'update timestamp
            Me._Timestamp = Now
        End If
        observation.ReleaseWriterLock()

    End Sub

#End Region

End Class
