Imports UvARescue.Agent
Imports UvARescue.Math

Imports System.Math

Public MustInherit Class IcpScanMatcher
    Implements IScanMatcher

    'for convenience 
    Public Const PI_2 As Double = PI / 2

    'break conditions
    Protected Const MAX_ITERATIONS As Integer = 80
    Protected Const MIN_CORRESPONDENCIESPCT As Single = 0.4 '40%
    Protected Const RESULT_MAXCONSIDERED As Integer = 10 'max num of results to consider 
    Protected Const RESULT_CONVERGED As Integer = 8 'how many good results before converged

    'only results below the error threshold are considered
    Protected ERROR_THRESHOLD As Double = 0.01

    'maximum distance within which to detect correlations
    Private MAX_CORRELATIONDISTANCE As Double = 2000


    Protected Overridable Sub ApplyConfig(ByVal config As Tools.Config) Implements IScanMatcher.ApplyConfig
        Me.ERROR_THRESHOLD = Double.Parse(config.GetConfig("slam", "error-threshold", CStr(Me.ERROR_THRESHOLD)))
        Me.MAX_CORRELATIONDISTANCE = Double.Parse(config.GetConfig("slam", "correlation-maxdistance", CStr(Me.MAX_CORRELATIONDISTANCE)))
    End Sub



    Public Function Match(ByVal scan1 As ScanObservation, ByVal scan2 As ScanObservation, ByVal seed As Pose2D) As MatchResult Implements IScanMatcher.Match

        If scan1.Length < 2 OrElse scan2.Length < 2 Then _
            Throw New InvalidOperationException("Cannot match scans with fewer than 2 scanpoints")

        Dim points1() As Vector2
        Dim points2() As Vector2

        If scan1.FieldOfView = scan2.FieldOfView AndAlso scan1.Resolution = scan2.Resolution Then
            'same scanner setup, reuse angles to save redundant computation
            Dim angles() As Double = Me.ComputeScanAngles(scan1)
            points1 = Me.ToPoints(scan1, angles)
            points2 = Me.ToPoints(scan2, angles)
        Else
            'differing scanner setup, compute angles for each scan individually
            points1 = Me.ToPoints(scan1, Me.ComputeScanAngles(scan1))
            points2 = Me.ToPoints(scan2, Me.ComputeScanAngles(scan2))
        End If


        Dim filter1() As Boolean = Me.ToFilter(scan1)
        Dim filter2() As Boolean = Me.ToFilter(scan2)
        Dim dangle As Double = scan1.FieldOfView / scan1.Length

        Return Me.MatchPoints(points1, points2, seed, dangle, filter1, filter2)

    End Function


    Protected MustOverride Function MatchPoints(ByVal points1() As Vector2, ByVal points2() As Vector2, ByVal odometry As Pose2D, ByVal dangle As Double, ByVal filter1() As Boolean, ByVal filter2() As Boolean) As MatchResult


    Protected Overridable Function HasConverged(ByVal numPoints As Integer, ByVal numCorrespondencies As Integer, ByVal numIterations As Integer) As Boolean
        Return numIterations < MAX_ITERATIONS _
            AndAlso numCorrespondencies > MIN_CORRESPONDENCIESPCT * numPoints
    End Function



    Protected Overridable Function ToPoints(ByVal scan As ScanObservation, ByVal angles() As Double) As Vector2()
        With scan
            Dim points(.Range.Length - 1) As Vector2
            'Dim xoffset As Double = scan.OffsetX * .Factor
            'Dim yoffset As Double = scan.OffsetY * .Factor

            Dim dist As Double, angle As Double
            For i As Integer = 0 To .Range.Length - 1
                dist = .Range(i) * .Factor
                angle = angles(i)
                points(i) = New Vector2(Cos(angle) * dist, Sin(angle) * dist)
            Next

            Return points
        End With
    End Function

    Protected Overridable Function ToFilter(ByVal scan As ScanObservation) As Boolean()
        With scan
            Dim filter(.Range.Length - 1) As Boolean
            For i As Integer = 0 To .Range.Length - 1
                filter(i) = .Range(i) < .MinRange OrElse .Range(i) > .MaxRange
            Next
            Return filter
        End With
    End Function

    Protected Overridable Function ToLocal(ByVal points() As Vector2, ByVal target As Pose2D) As Vector2()
        Dim rotmx As TMatrix2D = target.ToGlobalMatrix

        Dim local(points.Length - 1) As Vector2
        For i As Integer = 0 To points.Length - 1
            local(i) = rotmx * points(i)
        Next
        Return local

    End Function



    Protected Overridable Function ComputeScanAngles(ByVal scan As ScanObservation) As Double()

        Dim fov As Double = scan.FieldOfView
        Dim res As Double = scan.Resolution

        Dim length As Integer = CInt(fov / res) + 1
        Dim fromAngle As Double = -fov / 2

        Dim angles(length - 1) As Double
        For i As Integer = 0 To length - 1
            angles(i) = fromAngle + res * i
        Next
        Return angles

    End Function

    Protected Overridable Function CorrelatePoints(ByVal points1() As Vector2, ByVal points2() As Vector2, ByVal filter1() As Boolean, ByVal filter2() As Boolean) As Integer(,)

        Dim maxdist As Double = Pow(MAX_CORRELATIONDISTANCE, 2)

        'dists will hold the full distance matrix
        Dim dists(points1.Length - 1, points2.Length - 1) As Double

        '- collect preferences for points1
        '- and simultaneously construct full distance matrix dists
        Dim marks1to2(points1.Length - 1) As Integer
        For i As Integer = 0 To points1.Length - 1

            Dim curdist As Double = maxdist + 1
            Dim curmark As Integer = -1

            For j As Integer = 0 To points2.Length - 1

                'compute distance and store in full distance matrix
                Dim dist As Double = Me.ComputeSquaredDistance(points1(i), points2(j))
                dists(i, j) = dist

                'check preference
                If Not filter1(i) AndAlso Not filter2(j) _
                    AndAlso dist < maxdist AndAlso dist < curdist Then

                    curdist = dists(i, j)
                    curmark = j

                End If

            Next

            'store preference
            marks1to2(i) = curmark

        Next

        'matches will get to store the number of pairs that prefer each other
        Dim matches As Integer = 0

        'collect preferences for points2
        Dim marks2to1(points2.Length - 1) As Integer
        For j As Integer = 0 To points2.Length - 1

            Dim curdist As Double = maxdist + 1
            Dim curmark As Integer = -1

            'find nearest point of those that prefer me
            For i As Integer = 0 To points1.Length - 1
                If marks1to2(i) = j Then
                    Dim dist As Double = dists(i, j)
                    If dist < maxdist AndAlso dist < curdist Then
                        curdist = dist
                        curmark = i
                    End If
                End If
            Next

            marks2to1(j) = curmark

            If curdist < maxdist Then
                matches += 1
            End If

        Next

        'collect all pairs
        Dim pairs(matches - 1, 1) As Integer
        Dim p As Integer = 0
        For j As Integer = 0 To points2.Length - 1

            If marks2to1(j) < 0 Then
                Continue For
            End If

            pairs(p, 0) = marks2to1(j)
            pairs(p, 1) = j

            p += 1

        Next

        Return pairs

    End Function

    Protected Overridable Function ComputeSquaredDistance(ByVal point1 As Vector2, ByVal point2 As Vector2) As Double
        Return Pow(point2.X - point1.X, 2) + Pow(point2.Y - point1.Y, 2)
    End Function

End Class
