#region MathNet SignalProcessing, Copyright 2004 Christoph Ruegg

// MathNet SignalProcessing, part of MathNet
//
// Copyright (c) 2004,	Christoph Ruegg, http://www.cdrnet.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published 
// by the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#endregion

using System;

using MathNet.Numerics;

namespace MathNet.SignalProcessing.Filter
{
	/// <summary>
	/// Finite Impulse Response (FIR) Filters are based on
	/// fourier series and implemented using a discrete
	/// convolution equation. FIR Filters are always
	/// online, stable and causal. 
	/// </summary>
	public class FirFilter : IOnlineFilter
	{
		private Complex[] coefficients;
		private ShiftBuffer buffer;

		#region Constructors
		/// <summary>
		/// Finite Impulse Response (FIR) Filter.
		/// </summary>
		public FirFilter(FilterCoefficients coefficients)
		{
			this.coefficients = coefficients.Coefficients;
			this.buffer = new ShiftBuffer(this.coefficients.Length);
		}
		/// <summary>
		/// Finite Impulse Response (FIR) Filter.
		/// </summary>
		public FirFilter(Complex[] coefficients)
		{
			this.coefficients = coefficients;
			this.buffer = new ShiftBuffer(this.coefficients.Length);
		}
		#endregion

		/// <summary>
		/// Process a single sample.
		/// </summary>
		public Complex ProcessSamples(Complex sample)
		{
			buffer.ShiftAddHead(sample);
			return buffer.MultiplyAccumulate(coefficients);
		}

		/// <summary>
		/// Process a whole set of samples at once.
		/// </summary>
		public Complex[] ProcessSamples(Complex[] samples)
		{
			Complex[] ret = new Complex[samples.Length];
			for(int i=0;i<samples.Length;i++)
				ret[i] = ProcessSamples(samples[i]);
			return ret;
		}

		/// <summary>
		/// Reset internal state (not coefficients!).
		/// </summary>
		public void Reset()
		{
			buffer.Reset();
		}
	}
}
