#region MathNet SignalProcessing, Copyright 2004 Christoph Ruegg

// MathNet SignalProcessing, part of MathNet
//
// Copyright (c) 2004,	Christoph Ruegg, http://www.cdrnet.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published 
// by the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#endregion

using System;

using MathNet.Numerics;

namespace MathNet.SignalProcessing.Filter
{
	/// <summary>
	/// Infinite Impulse Response (FIR) Filters need much
	/// less coefficients (and are thus much faster) than
	/// comparable FIR Filters, but are potentially unstable.
	/// IIR Filters are always online and causal. This IIR
	/// Filter implements the canonic Direct Form II structure.
	/// </summary>
	public class IirFilter : IOnlineFilter
	{
		private Complex[] coefficients;
		private ShiftBuffer buffer;
		private int offset;

		#region Constructors
		/// <summary>
		/// Infinite Impulse Response (IIR) Filter.
		/// </summary>
		public IirFilter(FilterCoefficients coefficients)
		{
			this.coefficients = coefficients.Coefficients;
			if((this.coefficients.Length&1)!=0)
				throw new ArgumentException("Even number of coefficients required","coefficients");
			this.buffer = new ShiftBuffer(this.coefficients.Length);
			offset = this.coefficients.Length>>1;
		}
		/// <summary>
		/// Infinite Impulse Response (IIR) Filter.
		/// </summary>
		public IirFilter(Complex[] coefficients)
		{
			this.coefficients = coefficients;
			if((this.coefficients.Length&1)!=0)
				throw new ArgumentException("Even number of coefficients required","coefficients");
			this.buffer = new ShiftBuffer(this.coefficients.Length);
			offset = this.coefficients.Length>>1;
		}
		#endregion

		/// <summary>
		/// Process a single sample.
		/// </summary>
		public Complex ProcessSamples(Complex sample)
		{
			Complex un = coefficients[0]*sample + buffer.MultiplyAccumulate(coefficients,1,offset-1);
			buffer.ShiftAddHead(un);
			Complex yn = buffer.MultiplyAccumulate(coefficients,offset,offset);
			return yn;
		}

		/// <summary>
		/// Process a whole set of samples at once.
		/// </summary>
		public Complex[] ProcessSamples(Complex[] samples)
		{
			Complex[] ret = new Complex[samples.Length];
			for(int i=0;i<samples.Length;i++)
				ret[i] = ProcessSamples(samples[i]);
			return ret;
		}

		/// <summary>
		/// Reset internal state (not coefficients!).
		/// </summary>
		public void Reset()
		{
			buffer.Reset();
		}
	}
}
