Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.IO
Imports System.Windows.Forms

Imports UvARescue.Agent
Imports UvARescue.Tools

Public Class SaveMapDialog

    Private _Image As ManifoldImage
    Private _GeoRef As GeoReference

    Public Sub New(ByVal image As ManifoldImage)

        ' This call is required by the Windows Form Designer.
        InitializeComponent()

        ' Add any initialization after the InitializeComponent() call.
        Me._Image = image
        If Me._Image.GetGeoReferenceData(Me._GeoRef) Then
            With Me._GeoRef
                Me.txtOffsetX.Text = CStr(.OffsetX)
                Me.txtOffsetY.Text = CStr(.OffsetY)
                Me.txtPixelWidth.Text = CStr(.ScaleX)
                Me.txtPixelHeight.Text = CStr(.ScaleY)
                Me.txtShearingX.Text = CStr(.ShearX)
                Me.txtShearingY.Text = CStr(.ShearY)
            End With

            'default to OK
            Me.DialogResult = Windows.Forms.DialogResult.OK

        Else
            MsgBox("Could not load GeoReference data for Saving from ManifoldImage", MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
            Me.DialogResult = Windows.Forms.DialogResult.Cancel
            Me.Close()

        End If

    End Sub

    Private Sub btnSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSave.Click
        Try
            Me.Save()
        Catch ex As Exception
            MsgBox("Error occurred during Save" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
        End Try
    End Sub

    Private Sub btnCancel_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCancel.Click
        Me.Close()
    End Sub

    Private Sub Save()


        'get all supported save options
        Dim optvalues As Array = System.Enum.GetValues(GetType(ManifoldImage.SaveOption))
        Dim optnames() As String = System.Enum.GetNames(GetType(ManifoldImage.SaveOption))



        'prompt the user to specify folder to save images to
        Dim folder As String = ""
        With Me.dlgFolderBrowser
            Select Case .ShowDialog
                Case Windows.Forms.DialogResult.OK
                    folder = .SelectedPath
                Case Else
                    Exit Sub
            End Select
        End With


        'inspect selected image formats
        Dim formats As New List(Of ImageFormat)
        For Each item As ListViewItem In Me.lvwFormats.Items
            If item.Checked Then
                formats.Add(Me.GetImageFormat(CStr(item.Tag)))
            End If
        Next

        'to display progress, count the number of tasks
        Dim tasks As Integer = formats.Count
        tasks *= optvalues.Length 'each option for every format
        tasks *= 2 'world file generation for each format+save option
        If Me.chkGeoTiff.Checked Then
            tasks += 1 'GeoTIFF generation
        End If

        'setup progressbar
        With Me.barProgress
            .Value = 0
            .Minimum = 0
            .Maximum = tasks
        End With

        'construct base filename
        Dim basefilename As String = Me.txtName.Text.Trim
        If String.IsNullOrEmpty(basefilename) Then
            Throw New InvalidOperationException("No filname specified")
        Else
            'combine with selected folder
            basefilename = Path.Combine(folder, Path.GetFileNameWithoutExtension(basefilename))
        End If




        'now save, loop through formats
        For Each format As ImageFormat In formats

            'for each save option
            For i As Integer = 0 To optvalues.Length - 1

                'get save option
                Dim optvalue As ManifoldImage.SaveOption = CType(optvalues.GetValue(i), ManifoldImage.SaveOption)

                'construct filenames for this layer
                Dim imgFilename As String = String.Format("{0}-{1}-[{2}].{3}", basefilename, CInt(optvalue), optnames(i).ToLower, Me.GetImageFileExtension(format))
                Dim wldFilename As String = String.Format("{0}-{1}-[{2}].{3}", basefilename, CInt(optvalue), optnames(i).ToLower, Me.GetWorldFileExtension(format))

                'save the world file
                Me._GeoRef.SaveWorldFile(wldFilename)
                Me.barProgress.Value += 1


                Dim bitmap As Bitmap = Nothing
                If Me._Image.GetImageFileSaveData(bitmap, optvalue) Then
                    Try
                        bitmap.Save(imgFilename, format)
                    Catch ex As Exception
                        MsgBox("Could not Save as " & format.ToString.ToUpper & " to " & imgFilename & vbNewLine & ex.Message, MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
                    Finally
                        If Not IsNothing(bitmap) Then
                            bitmap.Dispose()
                        End If
                    End Try
                Else
                    MsgBox("Could not get data from Manifold for layer " & optnames(i), MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
                End If
                Me.barProgress.Value += 1

            Next

        Next

        If Me.chkGeoTiff.Checked Then
            Dim bitmap As Bitmap = Nothing
            If Me._Image.GetImageFileSaveData(bitmap, ManifoldImage.SaveOption.Map) Then
                Try
                    GeoTiff.Save(basefilename & Me.GetImageFileExtension(ImageFormat.Tiff), bitmap, Me._GeoRef)
                Catch ex As Exception
                    MsgBox("Could not Save as GeoTIFF" & vbNewLine & ex.Message, MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
                End Try
            Else
                MsgBox("Could not get data from Manifold for GeoTIFF", MsgBoxStyle.OkOnly Or MsgBoxStyle.Exclamation)
            End If
            Me.barProgress.Value += 1
        End If

        MsgBox("Map was saved successfully", MsgBoxStyle.OkOnly Or MsgBoxStyle.Information)

    End Sub

    Private Function GetImageFormat(ByVal extension As String) As ImageFormat
        Select Case extension.ToLower
            Case "tif", "tiff"
                Return ImageFormat.Tiff
            Case "gif"
                Return ImageFormat.Gif
            Case "jpg", "jpe", "jpeg"
                Return ImageFormat.Jpeg
            Case "png"
                Return ImageFormat.Png
            Case Else
                Return ImageFormat.Bmp
        End Select
    End Function

    Private Function GetImageFileExtension(ByVal format As ImageFormat) As String
        If format.Equals(ImageFormat.Tiff) Then
            Return "tif"
        ElseIf format.Equals(ImageFormat.Gif) Then
            Return "gif"
        ElseIf format.Equals(ImageFormat.Jpeg) Then
            Return "jpg"
        ElseIf format.Equals(ImageFormat.Png) Then
            Return "png"
        Else
            Return "bmp"
        End If
    End Function

    Private Function GetWorldFileExtension(ByVal format As ImageFormat) As String
        Dim imExt As String = Me.GetImageFileExtension(format)
        Dim wfExt As String = ""
        If imExt.Length > 2 Then
            'first char of imgext + last char of imgext + w 
            wfExt &= imExt.Substring(0, 1)
            wfExt &= imExt.Substring(imExt.Length - 1, 1)
            wfExt &= "w"
        Else
            'default to TFW
            wfExt = "tfw"
        End If
        Return wfExt
    End Function

End Class