Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports System.Math
Imports System.Windows.Forms

Imports UvARescue.Agent

Public Class ManifoldView
    Inherits UserControl
    Implements IManifoldView

#Region " Constructor / Destructor "

    Public Sub New(ByVal image As ManifoldImage)

        Me.InitializeComponent()

        Me.SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        Me.SetStyle(ControlStyles.OptimizedDoubleBuffer, True)
        Me.SetStyle(ControlStyles.UserPaint, True)

        Me._ShowAprioriLayer = True
        Me._ShowFreeSpaceLayer = True
        Me._ShowClearSpaceLayer = True
        Me._ShowObservationsLayer = True
        Me._ShowAgentsLayer = True

        Me.Image = image

    End Sub

    Private Sub InitializeComponent()
        Me.SuspendLayout()
        '
        'ManifoldView
        '
        Me.Name = "ManifoldView"
        Me.Size = New System.Drawing.Size(286, 201)
        Me.ResumeLayout(False)

    End Sub

    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If disposing Then
            If Not IsNothing(Me.Image) Then
                Me.Image.Dispose()
            End If
        End If
        MyBase.Dispose(disposing)
    End Sub

#End Region

#Region " Observe ManifoldImage "

    Private _Image As ManifoldImage
    Public Property Image() As ManifoldImage
        Get
            Return Me._Image
        End Get
        Set(ByVal value As ManifoldImage)
            If Not value Is Me._Image Then
                If Not IsNothing(Me._Image) Then
                    Me._Image.DetachView(Me)
                End If
                If Not IsNothing(value) Then
                    value.AttachView(Me)
                End If
                Me._Image = value
                Me.Invalidate()
            End If
        End Set
    End Property

    Public Sub NotifyManifoldImageUpdated() Implements Agent.IManifoldView.NotifyManifoldImageUpdated
        If Me.InvokeRequired Then
            Dim handler As New MethodInvoker(AddressOf Me.NotifyManifoldImageUpdated)
            Me.Invoke(handler)
        Else
            Me.Invalidate()
        End If
    End Sub

#End Region

#Region " Zooming and Panning "

    Private _AutoZoom As Boolean = True
    Private _AutoPan As Boolean = True
    Private _Zoom As Single = 1.0
    Private _Pan As New Point(0, 0)


    Public Sub ZoomIn()
        Me.ZoomTo(Me._Zoom + 0.1F)
    End Sub
    Public Sub ZoomOut()
        Me.ZoomTo(Me._Zoom - 0.1F)
    End Sub
    Public Sub ZoomTo(ByVal zoom As Single)
        Me._AutoZoom = False
        Me._Zoom = Max(zoom, 0.1F)
        Me.Invalidate()
    End Sub
    Public Sub AutoZoom()
        Me._AutoZoom = True
        Me._Zoom = 1.0
        Me._AutoPan = True
        Me._Pan = New Point(0, 0)
        Me.Invalidate()
    End Sub

    Protected Overrides Sub OnMouseWheel(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseWheel(e)
        Me.ZoomTo(Me._Zoom + DirectCast(IIf(e.Delta > 0, 0.1F, -0.1F), Single))
    End Sub


    Private isPanning As Boolean = False
    Private lastPoint As Point

    Protected Overrides Sub OnMouseDown(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseDown(e)
        If e.Button = Windows.Forms.MouseButtons.Left Then

            'start Panning
            Windows.Forms.Cursor.Current = Cursors.SizeAll
            Me._AutoPan = False
            Me.lastPoint = e.Location
            Me.isPanning = True

        ElseIf e.Button = Windows.Forms.MouseButtons.Right Then
            'reset 
            Me.AutoZoom()
        End If
    End Sub
    Protected Overrides Sub OnMouseMove(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseMove(e)
        If Me.isPanning Then
            Me._Pan.X += (e.X - Me.lastPoint.X)
            Me._Pan.Y += (e.Y - Me.lastPoint.Y)
            Me.lastPoint = e.Location
            Me.Invalidate()
        End If
    End Sub
    Protected Overrides Sub OnMouseUp(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseUp(e)
        If e.Button = Windows.Forms.MouseButtons.Left Then
            Me.isPanning = False
            Windows.Forms.Cursor.Current = Cursors.Default
        End If
    End Sub


#End Region

#Region " Layer Flags and Paint routine "

    Private _ShowAprioriLayer As Boolean
    Public Property ShowAprioriLayer() As Boolean
        Get
            Return Me._ShowAprioriLayer
        End Get
        Set(ByVal value As Boolean)
            If Not value = Me._ShowAprioriLayer Then
                Me._ShowAprioriLayer = value
                Me.Invalidate()
            End If
        End Set
    End Property

    Private _ShowFreeSpaceLayer As Boolean
    Public Property ShowFreeSpaceLayer() As Boolean
        Get
            Return Me._ShowFreeSpaceLayer
        End Get
        Set(ByVal value As Boolean)
            If Not value = Me._ShowFreeSpaceLayer Then
                Me._ShowFreeSpaceLayer = value
                Me.Invalidate()
            End If
        End Set
    End Property

    Private _ShowClearSpaceLayer As Boolean
    Public Property ShowClearSpaceLayer() As Boolean
        Get
            Return Me._ShowClearSpaceLayer
        End Get
        Set(ByVal value As Boolean)
            If Not value = Me._ShowClearSpaceLayer Then
                Me._ShowClearSpaceLayer = value
                Me.Invalidate()
            End If
        End Set
    End Property

    Private _ShowObservationsLayer As Boolean
    Public Property ShowObservationsLayer() As Boolean
        Get
            Return Me._ShowObservationsLayer
        End Get
        Set(ByVal value As Boolean)
            If Not value = Me._ShowObservationsLayer Then
                Me._ShowObservationsLayer = value
                Me.Invalidate()
            End If
        End Set
    End Property

    Private _ShowAgentsLayer As Boolean
    Public Property ShowAgentsLayer() As Boolean
        Get
            Return Me._ShowAgentsLayer
        End Get
        Set(ByVal value As Boolean)
            If Not value = Me._ShowAgentsLayer Then
                Me._ShowAgentsLayer = value
                Me.Invalidate()
            End If
        End Set
    End Property


    Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        MyBase.OnPaint(e)

        'renders the image of the manifold:
        '- scaled to fit the client area
        '- while preserving the aspect-ration (so equal scaling in both dimensions)
        '- and centered vertically nad horizontally

        If Not IsNothing(Me.Image) Then

            'compute the scaling of the image in both directions
            Dim xratio As Double = Me.Image.ImageWidth / Me.ClientSize.Width
            Dim yratio As Double = Me.Image.ImageHeight / Me.ClientSize.Height

            'compute the preferred scale to exactly fit the most 
            'restraining dimension
            Dim scale As Single
            If xratio > yratio Then
                'width is most restraining for current image bounds
                scale = CType(1 / xratio, Single)
            Else
                'height is most restraining for current image bounds
                scale = CType(1 / yratio, Single)
            End If

            If Not _AutoZoom Then
                'add the zooming factor that the user set with the scrollwheel
                scale *= Me._Zoom
            End If

            'ensure limits
            scale = Max(scale, 0.01F)


            'then compute the preferred offset that centers the 
            'image on screen given the current scale
            Dim offset As New Point
            offset.X = CInt((Me.ClientSize.Width - Me.Image.ImageWidth * scale) / 2 / scale)
            offset.Y = CInt((Me.ClientSize.Height - Me.Image.ImageHeight * scale) / 2 / scale)

            If Not Me._AutoPan Then
                'add panning
                offset.X += CInt(Me._Pan.X / scale)
                offset.Y += CInt(Me._Pan.Y / scale)
            End If

            Dim g As Graphics = e.Graphics
            g.PageUnit = GraphicsUnit.Pixel
            g.ScaleTransform(scale, scale)
            g.TranslateTransform(offset.X, offset.Y)

            Me.Image.Draw(g, ManifoldRenderingMode.GdiPlus, True, Me.ShowAprioriLayer, False, Me.ShowFreeSpaceLayer, Me.ShowClearSpaceLayer, Me.ShowObservationsLayer, True, Me.ShowObservationsLayer, Me.ShowAgentsLayer, True)

        End If

    End Sub

    Protected Overrides Sub OnResize(ByVal e As System.EventArgs)
        MyBase.OnResize(e)
        Me.Invalidate()
    End Sub

#End Region

End Class
